package net.shambolica.helperl;

import com.ericsson.otp.erlang.*;

import java.io.UnsupportedEncodingException;
import java.util.*;

import static net.shambolica.helperl.ET.propOrNull;

/**
 * Extraction of data from terms.
 */
public class EX {

    /*==========================================================
     * Proplists
     *=========================================================*/

    public static OtpErlangObject findInProplist(OtpErlangObject proplist, OtpErlangObject key) {

        if (proplist == null || ! (proplist instanceof OtpErlangList) || key == null) {
            return null;
        }

        for (OtpErlangObject entry : ((OtpErlangList)proplist).elements()) {
            OtpErlangTuple tuple = propOrNull(entry);
            if (tuple != null) {
                if (tuple.elementAt(0).equals(key)) {
                    return tuple.elementAt(1);
                }
            }
        }

        return null;

    }


    public static List<OtpErlangObject> findAllInProplist(OtpErlangObject proplist, OtpErlangObject key) {

        List<OtpErlangObject> result = new ArrayList<OtpErlangObject>();

        if (proplist != null && (proplist instanceof OtpErlangList) && key != null) {
            for (OtpErlangObject entry : ((OtpErlangList) proplist).elements()) {
                OtpErlangTuple tuple = propOrNull(entry);
                if (tuple != null) {
                    if (tuple.elementAt(0).equals(key)) {
                        result.add(tuple.elementAt(1));
                    }
                }
            }
        }

        return result;

    }


    /**
     * Extract all available values from proplist. The keys that will be scanned are:
     * <ul>
     *     <li>Atoms with the string value</li>
     *     <li>Binaries containing the UTF-8 encoded string value</li>
     *     <li>Lists containing the UTF-8 encoded string value</li>
     * </ul>
     * Note that ALL values will be returned.
     * @return an empty list in case of a non-proplist parameter.
     */
    public static List<OtpErlangObject> findAllInProplistByName(OtpErlangObject proplist, String key) {

        List<OtpErlangObject> result = new ArrayList<OtpErlangObject>();

        if (proplist != null && (proplist instanceof OtpErlangList) && key != null) {
            for (OtpErlangObject entry : ((OtpErlangList) proplist).elements()) {
                OtpErlangTuple tuple = propOrNull(entry);
                if (tuple != null) {
                    try {
                        if (key.equals(equivalentString(tuple.elementAt(0)))) {
                            result.add(tuple.elementAt(1));
                        }
                    } catch (OtpErlangRangeException e) {
                        // Ignore, skip onwards.
                    }
                }
            }
        }

        return result;

    }


    /**
     * Extract available values from proplist. The keys that will be scanned are:
     * <ul>
     *     <li>Atoms with the string value</li>
     *     <li>Binaries containing the UTF-8 encoded string value</li>
     *     <li>Lists containing the UTF-8 encoded string value</li>
     * </ul>
     * Note that only the first value will be returned.
     * @return an empty map in case of a non-proplist parameter.
     */
    public static Map<String, OtpErlangObject> findInProplistByNames(OtpErlangObject proplist, Collection<String> keys) {

        Map<String, OtpErlangObject> result = new LinkedHashMap<String, OtpErlangObject>();

        if (proplist != null && (proplist instanceof OtpErlangList)) {
            OtpErlangList list = (OtpErlangList) proplist;
            for (OtpErlangObject entry : list.elements()) {
                OtpErlangTuple tuple = propOrNull(entry);
                if (tuple != null) {
                    try {
                        String key = equivalentString(tuple.elementAt(0));
                        if (key != null && keys.contains(key) && !result.containsKey(key)) {
                            result.put(key, tuple.elementAt(1));
                        }
                    } catch (OtpErlangRangeException e) {
                        // Ignore, skip onwards.
                    }
                }
            }
        }

        return result;
    }


    /**
     * Returns the appropriate string contents of the term, or null if the term is not
     * of appropriate type.
     * @throws OtpErlangRangeException if values are outside expected norm, e.g. list containing non-byte values.
     */
    public static String equivalentString(OtpErlangObject oeo) throws OtpErlangRangeException {
        if (oeo == null) {
            return null;
        } else if (oeo instanceof OtpErlangAtom) {
            return ((OtpErlangAtom)oeo).atomValue();
        } else if (oeo instanceof OtpErlangString) {
            return ((OtpErlangString)oeo).stringValue();
        } else if (oeo instanceof OtpErlangBinary) {
            try {
                return new String(((OtpErlangBinary) oeo).binaryValue(), "UTF-8");
            } catch (UnsupportedEncodingException e) {
                throw new IllegalStateException("utf-8 encoding not found, ha");
            }
        } else if (oeo instanceof OtpErlangList) {
            OtpErlangList list = (OtpErlangList)oeo;
            byte[] bytes = new byte[list.arity()];
            int i = 0;
            for (OtpErlangObject element : list) {
                if (! (element instanceof OtpErlangLong)) {
                    throw new OtpErlangRangeException("non-integer value in string list");
                } else if (((OtpErlangLong)element).longValue() < 0 || ((OtpErlangLong)element).longValue() > 255) {
                    throw new OtpErlangRangeException("non-byte value in string list");
                } else {
                    bytes[i++] = ((OtpErlangLong)element).byteValue();
                }
            }
            try {
                return new String(bytes, "UTF-8");
            } catch (UnsupportedEncodingException e) {
                throw new IllegalStateException("utf-8 encoding not found, ha");
            }
        } else {
            return null;
        }

    }

}
