package net.shambolica.helperl.processes;

import com.ericsson.otp.erlang.*;
import net.shambolica.helperl.ET;

import static net.shambolica.helperl.ET.eAtom;
import static net.shambolica.helperl.ET.eTuple;
import static net.shambolica.helperl.util.DataUtils.isEmpty;
import static net.shambolica.helperl.util.DataUtils.trim;

/**
 * A simple named or un-named service which responds to the following protocol:
 * <ul>
 *     <li>Receive query: {'query', From :: pid(), Ref :: term(), Query :: term()}</li>
 *     <li>Send response: {'reponse', Ref, Response :: term()}</li>
 *     <li>Receive ping: {'ping', From :: pid(), Ref :: term()}</li>
 *     <li>Send pong: {'pong', Ref}</li>
 *     <li>Shutdown: 'stop'</li>
 * </ul>
 * The actual service implementation should be fast to respond, or be able to asynchronously
 * handle multiple queries.
 */
public abstract class ESimpleService implements Runnable {

    protected final OtpNode node;
    protected final OtpMbox mbox;
    protected final OtpMbox tx;
    protected final String name;

    private final Thread thread;
    private volatile boolean isRunning;


    private static final OtpErlangAtom stopAtom = eAtom("stop");
    private static final OtpErlangAtom requestAtom = eAtom("request");
    private static final OtpErlangAtom responseAtom = eAtom("response");
    private static final OtpErlangAtom pingAtom = eAtom("ping");
    private static final OtpErlangAtom pongAtom = eAtom("pong");


    protected ESimpleService(OtpNode node) {
        this(node, null);
    }


    protected ESimpleService(OtpNode node, String name) {

        name = trim(name);

        this.node = node;
        this.mbox = isEmpty(name) ? node.createMbox() : node.createMbox(name);
        this.tx = node.createMbox();
        this.name = name;

        this.isRunning = true;
        this.thread = new Thread(this);
        this.thread.setDaemon(true);
        this.thread.start();

    }


    protected class ESSRequest {

        public final OtpErlangPid from;
        public final OtpErlangObject ref;
        public final OtpErlangObject request;

        public ESSRequest(OtpErlangPid from, OtpErlangObject ref, OtpErlangObject request) {
            this.from = from;
            this.ref = ref;
            this.request = request;
        }

        public void respond(OtpErlangObject response) {
            sendResponse(from, ref, response);
        }

        @Override
        public String toString() {
            return "ESSRequest{" +
                    "from=" + from +
                    ", ref=" + ref +
                    ", request=" + request +
                    '}';
        }
    }


    public String getName() {
        return name;
    }


    public OtpErlangPid getPid() {
        return mbox.self();
    }


    public OtpErlangPid self() {
        return getPid();
    }


    protected void sendResponse(ESSRequest request, OtpErlangObject response) {
        sendResponse(request.from, request.ref, response);
    }


    protected void sendResponse(OtpErlangPid to, OtpErlangObject ref, OtpErlangObject response) {
        if (isRunning) {
            synchronized(tx) {
                tx.send(to, ET.eTuple(responseAtom, ref, response));
            }
        }
    }


    protected abstract void notifyReceived(ESSRequest request);


    protected void notifyBadMessage(OtpErlangObject msg) {
        // Do nothing.
    }


    protected void notifyExitSignal(Exception e) {
        // Do nothing.
    }


    protected void notifyStopOnError(Exception e) {
        // Do nothing.
    }


    protected void notifyStop() {
        // Do nothing.
    }


    @Override
    public void run() {
        while (isRunning) {
            try {

                OtpErlangObject msg;

                msg = mbox.receive();

                if ((msg instanceof OtpErlangAtom) && (((OtpErlangAtom)msg).atomValue().equals(stopAtom.atomValue()))) {
                    notifyStop();
                    isRunning = false;
                } else if (isPingMessage(msg)) {
                    replyToPingMessage(msg, mbox);
                } else if (!isRequestMessage(msg)) {
                    notifyBadMessage(msg);
                } else {
                    OtpErlangTuple tuple = (OtpErlangTuple)msg;
                    OtpErlangPid from = (OtpErlangPid)(tuple.elementAt(1));
                    OtpErlangObject ref = tuple.elementAt(2);
                    OtpErlangObject request = tuple.elementAt(3);
                    notifyReceived(new ESSRequest(from, ref, request));
                }

            } catch (OtpErlangDecodeException e) {
                notifyStopOnError(e);
                isRunning = false;
            } catch (OtpErlangExit e) {
                notifyExitSignal(e);
                isRunning = false;
            } catch (RuntimeException e) {
                notifyStopOnError(e);
                isRunning = false;
            }
        }
    }


    private boolean isRequestMessage(OtpErlangObject msg) {
        boolean isGoodMessage;OtpErlangTuple tuple;
        if (! (msg instanceof OtpErlangTuple)) {
            isGoodMessage = false;
        } else {
            tuple = (OtpErlangTuple) msg;
            if (tuple.arity() != 4) {
                isGoodMessage = false;
            } else if (! (tuple.elementAt(0) instanceof OtpErlangAtom)) {
                isGoodMessage = false;
            } else if (! requestAtom.equals(tuple.elementAt(0))) {
                isGoodMessage = false;
            } else if (! (tuple.elementAt(1) instanceof OtpErlangPid)) {
                isGoodMessage = false;
            } else {
                isGoodMessage = true;
            }
        }
        return isGoodMessage;
    }


    private boolean isPingMessage(OtpErlangObject msg) {
        return (msg instanceof OtpErlangTuple) &&
                (((OtpErlangTuple)msg).arity() == 3) &&
                (((OtpErlangTuple)msg).elementAt(0).equals(pingAtom));
    }


    private void replyToPingMessage(OtpErlangObject msg, OtpMbox mbox) {
        OtpErlangPid from = (OtpErlangPid) ((OtpErlangTuple)msg).elementAt(1);
        OtpErlangObject ref = ((OtpErlangTuple)msg).elementAt(2);
        mbox.send(from, eTuple(pongAtom, ref));
    }

}
